///////////////////////////////////////////////////////////////////////////////
// Tutorial AddIns.cs
// Copyright (C) 2012 GEO-SLOPE International, Ltd.  All rights reserved.
//
// This file contains examples of Add-In functions that illustrate how to define
// an Add-Ins as part of a GeoStudio simulation. 
//
// Notes:
// 	Any code following a // is a comment
// 
//	You can change the names of functions and variables EXCEPT the function 
//  called "Calculate(double x)" which MUST exist inside the Add-In Class.
// 
//	Variables you want to see in the KeyIn Function dialog should be "public" 
//  in the Class
//
//   You can generate an Add-In assembly by running the gsaddin.cmd script
//   provided with the GeoStudio Add-In Developers Kit. 
///////////////////////////////////////////////////////////////////////////////
using System;

///////////////////////////////////////////////////////////////////////////////
// Add-In Function 1 : SineCurve 
//
// This is the first Add-In Function Class in this file.  You can have more 
// than one.  The name is "SineCurve", which will be displayed in the KeyIn 
// Functions dialog when the AddIn assembly is selected.
//
// This function takes the x value of the function type and returns the sin(x)
// Initial_Amplitued, Maximum_Amlitued and Wavelength are all fields that can be
// assigned when the function is selected in s KeyIn Functions dialog.
///////////////////////////////////////////////////////////////////////////////

// NOTE: Both the class and Calculate method must be public if they are to be 
// seen by GeoStudio when scanning the Assembly.
public class SineCurve 
{
    // Here is the list of variables you want to enter using the KeyIn Function 
    // dialog. All public variables are seen as fields in the KeyIn Function 
    // dialog.
    public double Initial_Amplitude;
    public double Maximum_Amplitude;
    public double Wavelength;
 
    // this is the main function calculator
    public double Calculate( double x )  
    {
        double y;           // declare the variable to return to the solver
        double k,d;         // Working variables.

        k = 6.28318531 / Wavelength; 
        d = Initial_Amplitude / Maximum_Amplitude; // phase shift

        // calculate the value to return
        y = Math.Sin(k*x+d);

        y = y * Maximum_Amplitude;

        // return the function Y value
        return(y);
    }
}

///////////////////////////////////////////////////////////////////////////////
// Add-In Function 2 : ScaleByElevation
//
// This function returns the function value scaled by the elevation at a node. 
// This function is expected to be used in cases where it is applied to node, 
// for example as a part of a boundary condition.
//
// The elevation of the current node is requested using the GetParam method.
// An instance of this function is created for every node its applied to in 
// the model. Thus there is no need to specify the node when asking for the
// elevation, its always the current functions node. 
///////////////////////////////////////////////////////////////////////////////

// The : Gsi.Function text after the name is required to call GetParam for 
// "live" data from the solver. The allowable "live" data variables depend upon
// where this Function is used. See SDK documentation references section for a 
// complete list of solver parameters. 
public class ScaleByElevation : Gsi.Function
{
    // An optional parameter the user can specifiy for graphing this 
    // Function. GetParam can only be called withing the solver.
    public double Elevation_For_Viewing;

    public double Calculate(double x)
    {
        double elevation;
        try
        {
            // GetParam only works in the solver
            elevation = GetParam(Gsi.DataParamType.eElevation);
        }
        catch
        {
            // Allow function to work so it can be graphed in GeoStudio.
            elevation = Elevation_For_Viewing;
        }

        return elevation * x;
    }
}


