// Sample Add-In BC Conduuctivity functions. 
//
// Copyright (C) 2012 GEO-SLOPE International, Ltd.  All rights reserved.
using System;

// There are two general functions that get used within other functions. 
// These are not called directly by the solvers.  They are used in the 
// other functions below.
public class My_General_Functions 
{
	// This is a C# random number generation function
	// It is used in other functions within this file.

	public static Random autoRand = new Random(); 

	//This first function takes a pressure and returns the Van_G K value 
	// given the hard wired a,n,m and Ksat values.
	// It is a general function here because the same code is used
	// twice below and it is not necessary repeating it both times
	// if it is made public to all other function in this file.
	// Notice that it does NOT have a Calculate() method in it.  It will not
	// show up in DEFINE.  It will just be available to other function in this file.
	
	public static double Van_G_K_Unsat( double pressure, double fa, double fn, double fm, 
		double fKsat ) 
	{
		
		// returned K value
		double fKx;	

		// temporary variables
		double fTemp1, fTemp2, fTemp3, fTemp4, fTemp5, fTemp6;

		if(pressure < 0.0) // if in the unsaturated side of the function
		{
			double fSuction = Math.Abs (pressure);
			fTemp1 = fSuction*fa;
			fTemp2 = (Math.Pow((1.0 + Math.Pow(fTemp1, fn)), (fm/2)));
			fTemp3 = Math.Pow(fTemp1, (fn-1));
			fTemp4 = (1.0 + Math.Pow(fTemp1, fn));
			fTemp5 = Math.Pow(fTemp4, -fm);
			fTemp6 = Math.Pow((1.0 - fTemp3 * fTemp5), 2.0);
			fKx = fKsat * (fTemp6/fTemp2);  
		}
		else // use the user input Ksat if pwp are zero or positive
			fKx = fKsat;		
		 
		return fKx;
	}

	// This is the second general function in this file.  It is called by other functions 
	// below that do have a Calculate() method in them.

	public static double Van_G_VWC( double pressure, double fa, double fn, double fm, 
		double fPorosity, double fResidualWC ) 
	{
		double fWC, suction;	// returned K value
		double fTemp1, fTemp2; // temporary variables

		if(pressure < 0.0) // if in the unsaturated side of the function
		{
			suction = Math.Abs (pressure);
			fTemp1 = suction*fa;
			fTemp2 = Math.Pow(  1.0 /  (1.0 + Math.Pow(fTemp1, fn)  ) , fm );
			fWC = fResidualWC + (fPorosity-fResidualWC) * fTemp2;  
		}
		else // use the user input porosity if pwp are zero or positive
			fWC = fPorosity;		
		 
		return fWC;
	}
}  // end of general functions in this file

//**************************************************************************************************
// This function will let the Ksat of a conductivity function
// vary randomly one order of magnitude across the profile.
// It DOES have a Calculate() method so will be visible
// as an Add-In function in DEFINE.

public class Random_Van_G_K_Unsat : Gsi.Function
{
	public double Ksat;		
	public double alpha;	// Van G "a" parameter in units of 1/pressure 
	public double n;			// Van G "n" parameter, unitless
	public double m;		// Van G "m" parameter, unitless
	double KRandomOffset;	

	// this is called a constructor.  It will set values the first
	// time each instance of this function is called.
	// Here, it assigns a random number that will remain the same
	// throughout the analysis.

	public Random_Van_G_K_Unsat()
	{
		// returns a double between 0.0 and 1.0
		KRandomOffset = My_General_Functions.autoRand.NextDouble(); 
	}

	public double Calculate( double pressure )
	{
		
		double fKx = My_General_Functions.Van_G_K_Unsat(pressure,alpha,n,m,Ksat);	
				 
		fKx = Math.Log10(fKx) ; 
				 
		// add up to 1 order of magnitude to K (which is Log at this point)
		if(KRandomOffset > 0.5) 
			fKx = fKx + KRandomOffset;
		else
			fKx = fKx - KRandomOffset; // make 1 order of magnitude less that K      

		return Math.Pow(10.0,fKx);
	}

}

//**************************************************************************************************

// sample function to return the unsat K function based on a Ksats as input by the user
public class Van_Genuchten_K_Unsat : Gsi.Function
{
	public double Ksat;		
	public double alpha_one_over_pressure;  // Van G "a" parameter in units of 1/pressure 
	public double n;			// Van G "n" parameter, unitless
	public double m;		// Van G "m" parameter, unitless

	public double Calculate( double pressure )
	{
	
		double fKx = My_General_Functions.Van_G_K_Unsat(pressure,alpha_one_over_pressure,n,m,Ksat);	
	
		return fKx;
	}
}

//**************************************************************************************************

public class Van_Genuchten_VWC : Gsi.Function
{
	public double Porosity;		
	public double alpha_one_over_pressure;  // Van G "a" parameter in units of 1/pressure 
	public double n;			// Van G "n" parameter, unitless
	public double m;		// Van G "m" parameter, unitless
	public double Residual_WC;

	// constructor to initialize data so graph does not give an error in DEFINE
	public Van_Genuchten_VWC()  
	{
		Porosity = 0.5;		
		alpha_one_over_pressure = 0.1 ; // Van G "a" parameter in units of 1/pressure 
		n = 1.5 ;			// Van G "n" parameter, unitless
		m = 0.5 ;			// Van G "m" parameter, unitless
		Residual_WC = 0.05 ;
	}

	public double Calculate( double pressure )
	{
		double fWC = My_General_Functions.Van_G_VWC(pressure, 
			alpha_one_over_pressure, n, m, Porosity, Residual_WC);	
		return fWC;
	}
}
