// SIGMA/W Add-In Constitutive Model.  
//
// Written by Greg Newman.
// Copyright (C) 2012 GEO-SLOPE International, Ltd.  All rights reserved.
using System;

//  This is the Linear elastic model.  It has the CalculateMatrix and UpdateStresses methods so is 
//  a valid  Add-In constitutive model.

//  This file also has a third, optional function that fills in the 4x4 [C] matrix.  It was made an 
//  optional function because it is needed twice.. Once in the formation of the [C] matrix for setting
//  up the equations, and once in the calculation of stresses after the equations are formed.  There 
//  is no point writing the same code twice if you can make it a function once and then use it in 
//  more than one place.
public class Linear_Elastic_Model : Gsi.Function
{
    // constants fields for user input
    public double E_Modulus;
    public double Poisson_Ratio;

    public void CalculateMatrix(Gsi.Matrix mCee)
    {
        CalcElasticCee(E_Modulus, Poisson_Ratio, mCee);
    }

    public void UpdateStresses(Gsi.Matrix mIncStrain, ref Gsi.Stresses mStresses)
    {
        // computed herein using Cee and Strains		
        Gsi.Matrix mIncStress = Gsi.Matrix.Zero(1, 4);
        Gsi.Matrix mCee = Gsi.Matrix.Zero(4, 4);

        // compute elastic Ce used to know the stress increment
        CalcElasticCee(E_Modulus, Poisson_Ratio, mCee);

        // {IncStress} = [Ce] x {IncStrain}
        mIncStress = mIncStrain * mCee;

        mStresses.x = GetParam(Gsi.DataParamType.eXTotalStress) + mIncStress[0, 0];
        mStresses.y = GetParam(Gsi.DataParamType.eYTotalStress) + mIncStress[0, 1];
        mStresses.z = GetParam(Gsi.DataParamType.eZTotalStress) + mIncStress[0, 2];
        mStresses.xyShear = GetParam(Gsi.DataParamType.eXYShearStress) + mIncStress[0, 3];

        mStresses.voidRatio = 0.0;   //   This is an optional return parameter if it is known.
    }

    //  This function is local to this code and returns a 4x4 matrix for a gauss point with elastic
    //  properties only.  It takes a modulus and poisson's ratio as input.  It is used by the main 
    //  two functions above.

    public static void CalcElasticCee(double fE, double fPoisson, Gsi.Matrix mCee)
    {
        double fCOM = fE / ((1.0 + fPoisson) * (1.0 - 2.0 * fPoisson));
        double fCOM1 = 1.0 - fPoisson;
        mCee[0, 0] = fCOM * fCOM1;
        mCee[0, 1] = fCOM * fPoisson;
        mCee[0, 2] = fCOM * fPoisson;
        mCee[0, 3] = 0.0;

        mCee[1, 0] = fCOM * fPoisson;
        mCee[1, 1] = fCOM * fCOM1;
        mCee[1, 2] = fCOM * fPoisson;
        mCee[1, 3] = 0.0;

        mCee[2, 0] = fCOM * fPoisson;
        mCee[2, 1] = fCOM * fPoisson;
        mCee[2, 2] = fCOM * fCOM1;
        mCee[2, 3] = 0.0;

        mCee[3, 0] = 0.0;
        mCee[3, 1] = 0.0;
        mCee[3, 2] = 0.0;
        mCee[3, 3] = 0.5 * fE / (1.0 + fPoisson);
    }
}  // this is the end of the main class in this file.
