// Sample Add-In a Von Mises yield criteria elastic plastic model. 
//
// Copyright (C) 2012 GEO-SLOPE International, Ltd.  All rights reserved.
using System;


// This is a Von Mises yield criteria elastic plastic model.  It HAS the CalculateMatrix and 
// UpdateStresses methods which MUST be present in an Add-In constitutive model.  .

public class Von_Mises_Model : Gsi.Function
{
	// declare global constants, these will appear in the Key In Materials Dialogue 
	//for data entry
	public double Cu; 
	public double E;
	public double Poisson;

	// These are used locally in this code and dont require input by the user
	double fLastYield;
	double PWP;

	// There are four local methods called inside this Von Mises model.  They are used
	// as part of the main two methods that required in the Add-In model.  Their usage
	// in the code below is highlighted in blue.

	// The first returns a 4x4 matrix for a gauss point with elastic properties only.  
	// It takes a modulus and poissons ratio as input.

	public void CalcElasticCee(double youngsModulus, double Poisson, 
		Gsi.Matrix mCee ) // returns mCee matrix
	{
		double fCOM  = youngsModulus / ((1.0 + Poisson)*(1.0 - 2.0*Poisson));
		double fCOM1 = 1.0 - Poisson;

		mCee[0,0] = fCOM*fCOM1;
		mCee[0,1] = fCOM*Poisson;
		mCee[0,2] = fCOM*Poisson;
		mCee[0,3] = 0.0;

		mCee[1,0] = fCOM*Poisson;
		mCee[1,1] = fCOM*fCOM1;
		mCee[1,2] = fCOM*Poisson;
		mCee[1,3] = 0.0;

		mCee[2,0] = fCOM*Poisson;
		mCee[2,1] = fCOM*Poisson;
		mCee[2,2] = fCOM*fCOM1;
		mCee[2,3] = 0.0;

		mCee[3,0] = 0.0;
		mCee[3,1] = 0.0;
		mCee[3,2] = 0.0;
		mCee[3,3] = 0.5*youngsModulus/(1.0+Poisson);
	}
 
	// This returns the stress invariants J2, J3 and Lode Angle and deviatoric stresses.  It 
	// takes as input x,y,z stress and shear stress..

	public void CalcStressInvariants(Gsi.Matrix mStresses, 
		out Gsi.Matrix mDevStress, out double  fJ2, out double  fJ3, 
		out double  fTheta, out double fSigMean )
	{
		double fRootJ2;
		double dSin3T;
		
		double fSigX = mStresses[0,0];
		double fSigY = mStresses[0,1];
		double fSigZ = mStresses[0,2];
		double fTauXY = mStresses[0,3];

		fSigMean = (fSigX+fSigY+fSigZ)/3.0;

		// deviatoric stresses
		mDevStress = mStresses - fSigMean;
		mDevStress[0,3] = mStresses[0,3];  // reset shear stress

		fJ2 = 0.16667 * (  Math.Pow(fSigX-fSigY,2.0) + Math.Pow(fSigY-fSigZ,2.0) + 
			Math.Pow(fSigZ-fSigX,2.0)  ) + Math.Pow(mDevStress[0,3],2.0); 

		fRootJ2 = Math.Pow(fJ2,0.5);
	
		double sx,sy,sz;

		sx = (2*fSigX - fSigY- fSigZ)/3.0;
		sy = (2*fSigY - fSigZ- fSigX)/3.0;
		sz = (2*fSigZ - fSigX- fSigY)/3.0;

		fJ3 = sx*sy*sz - sx*fTauXY*fTauXY;
		
		if(fRootJ2 == 0.0)
			dSin3T = 0.0;
		else
			dSin3T = -3.0*Math.Pow(3.0,0.5)*fJ3/(2.0*fJ2*fRootJ2);

		if(dSin3T >  1.0) 
			dSin3T =  1.0;
		if(dSin3T < -1.0) 
			dSin3T = -1.0;

		fTheta = Math.Asin(dSin3T) / 3.0;
	}

	
	// This returns the yield amount stress based on Von Mises.  It uses the 
	// stress invariant method coded above

	public void Von_mises_Yield(Gsi.Matrix mStresses, out double fYield)
	{
		Gsi.Matrix mDevStress = Gsi.Matrix.Zero(1,4);	// Temp Vector
		double fJ2, fJ3, fTheta, fSigMean;

		CalcStressInvariants(mStresses, out mDevStress, out fJ2, out fJ3, out fTheta, 
			out fSigMean);

		double fq = Math.Pow(3*fJ2,0.5); 
				
		// Check yield in the post iteration check stress algorithm	
		fYield = fq-Math.Pow(3,0.5)*Cu;
	}

	// This returns the Von_Mises plastic flow vector.  This also uses the stress
	// invariant method code above.

	public void Von_mises_PlasticFlow(  Gsi.Matrix mStresses, out Gsi.Matrix mPlasticFlow)  
	{
		double fJ2, fJ3, fTheta, fSigMean; 
		Double fTemp;

		Gsi.Matrix mDevStress = Gsi.Matrix.Zero(1,4);	// Temp Vector

		CalcStressInvariants(mStresses,out mDevStress,out fJ2,
			out fJ3, out fTheta, out fSigMean);
		
		fTemp = 0.5 / Math.Pow(fJ2,0.5); 
		mDevStress *= fTemp; 
		mDevStress[0,3] *= 2.0;  

		fTemp = Math.Pow(3.0,0.5);

		mPlasticFlow = mDevStress*fTemp; 
	}


	// These are the two REQUIRED methods that MUST exist in a user constitutive model.
	// The first returns the [C] matrix that will get put into the EPM at each iteration prior to 
	// SOLVING the equations
	
	// The second computes the new stress state after the incremental displacements and 
	// strains are computed in SOLVE at each iteration. 

	public void CalculateMatrix( Gsi.Matrix mCee )
	{     

		CalcElasticCee(E, Poisson, mCee); // compute elastic [C]

		int ItNum = (int)GetParam(Gsi.DataParamType.eStepIterationCount);	

		PWP = GetParam(Gsi.DataParamType.ePWP);
		double Sx = GetParam(Gsi.DataParamType.eXTotalStress) - PWP;
		double Sy = GetParam(Gsi.DataParamType.eYTotalStress) - PWP;
		double Sz = GetParam(Gsi.DataParamType.eZTotalStress) - PWP;
		double Sxy = GetParam(Gsi.DataParamType.eXYShearStress);

		Gsi.Matrix mCurrentStress = Gsi.Matrix.Zero(1,4);	
		// current solution effective stress state prior to adding the current 
		// incremental stress

		mCurrentStress[0,0] = Sx;
		mCurrentStress[0,1] = Sy;
		mCurrentStress[0,2] = Sz;
		mCurrentStress[0,3] = Sxy;

		if(ItNum > 1)
		{				
			if(fLastYield > 0.0)  
			{
				Gsi.Matrix mdFdSig = Gsi.Matrix.Zero(1,4);
				Gsi.Matrix mdGdSig = Gsi.Matrix.Zero(4,1);

				Gsi.Matrix mTempVRow = Gsi.Matrix.Zero(1,4);	
				Gsi.Matrix mTempVCol = Gsi.Matrix.Zero(4,1);	
				Gsi.Matrix mCep = Gsi.Matrix.Zero(4,4);	

				Von_mises_PlasticFlow(mCurrentStress, out mdFdSig);  

				mdGdSig = Gsi.Matrix.Transpose(mdFdSig);

				mTempVCol = mCee * mdGdSig  ; 
			
				mTempVRow = Gsi.Matrix.Transpose(mTempVCol);
				
				double fTemp = Gsi.Matrix.Dot(mdFdSig,mTempVCol);      
			       
				if(Math.Abs(fTemp)  < 1e-10) fTemp = 1.0e-10;

				fTemp = 1.0/fTemp;
    
				mCep = Gsi.Matrix.TensorCross(mTempVCol, mTempVRow);  
    
				mCep *= fTemp;

				//Form plastic constitutive matrix.
				mCee = mCee - mCep;  // [Cee-Cep]
			}
		}
	}

	public void UpdateStresses( Gsi.Matrix mIncStrain, ref Gsi.Stresses Stresses) 
	{     
		Gsi.Matrix mIncStress = Gsi.Matrix.Zero(1,4); 
		Gsi.Matrix mCee = Gsi.Matrix.Zero(4,4);	   
		Gsi.Matrix mStartStress = Gsi.Matrix.Zero(1,4);	
		Gsi.Matrix mUpdatedStress = Gsi.Matrix.Zero(1,4);	
		
		Gsi.Matrix mStressYield = Gsi.Matrix.Zero(1,4);	 
		Gsi.Matrix mExcessStress = Gsi.Matrix.Zero(1,4);	

		Gsi.Matrix mPlasticStress = Gsi.Matrix.Zero(4,1);
		Gsi.Matrix mPlasticStressT = Gsi.Matrix.Zero(1,4);	
		Gsi.Matrix mPlasticStrain = Gsi.Matrix.Zero(1,4);	
		Gsi.Matrix mPlasticStrainT = Gsi.Matrix.Zero(4,1);	
		Gsi.Matrix mdFdSig = Gsi.Matrix.Zero(1,4);  
		Gsi.Matrix mdGdSig = Gsi.Matrix.Zero(4,1);
		Gsi.Matrix mTempVec = Gsi.Matrix.Zero(1,4);	

		Double fYield = 0.0;										Double fYieldStart = 0.0;							
		Double fReductionFactor = 1.0;	
		Double fTemp = 0.0;								
		Double fDLambda;

		int i, iSubInc;
		iSubInc = 20;  // we will subdivide any yielded strains into smaller pieces
		// so that we can compute the excess stresses that need
		// to be subtracted from the elastic overshoot to get us 
		// back to the yield surface.

		int ItNum = (int)GetParam(Gsi.DataParamType.eStepIterationCount);

		// form the elastic Cee matrix as it is needed for stress calculations
		CalcElasticCee(E, Poisson, mCee); 

		// compute elastic Ce used to know the stress increment
		// {IncStress} = [Cee]x{IncStrain}   This is the full incremental stress 
		// based on the elastic load

		mIncStress = mIncStrain * mCee;  
		
		mStartStress[0,0] = GetParam(Gsi.DataParamType.eXTotalStress)   - PWP;
		mStartStress[0,1] = GetParam(Gsi.DataParamType.eYTotalStress)   - PWP;
		mStartStress[0,2] = GetParam(Gsi.DataParamType.eZTotalStress)   - PWP;
		mStartStress[0,3] = GetParam(Gsi.DataParamType.eXYShearStress) ;

		// compute updated stresses with full elastic increment applied. 
		// This will be reduced later if there has been yield
		mUpdatedStress = mStartStress + mIncStress;

		// check yield with starting stresses and then with updated stresses
		Von_mises_Yield(mStartStress,out fYieldStart);
		Von_mises_Yield(mUpdatedStress,out fYield);
	
		// find reduction factor
		if(fYield < 0.0)			// still elastic
			fReductionFactor = 0.0;
		else if(fYieldStart >= 0.0)	 // was plastic so stays plastic
			fReductionFactor = 1.0;
		else				// part elastic, part plastic
		{
			if(fYield*fYieldStart > 0) 
				if(fYield > 0)
					fReductionFactor = 1.0;
				else	
					fReductionFactor = 0.0;
			else            		
				fReductionFactor = fYield/(fYield-fYieldStart);
		}
	
		mStressYield = mStartStress + mIncStress*(1.0-fReductionFactor); 

		if( fReductionFactor > 0.0) 
			// there are some plastic strain induced stresses that need to be reduced
		{
			fLastYield = 1.0;			
			
			mExcessStress = mIncStress*fReductionFactor/iSubInc;

			// find plastic strain for the given total strain increment, compute the 
			// elastic stress that would have caused this strain, 
			// then reduce the initially computed stress increment by this plastic 
			//stress amount (there can be no change in stress for plastic strain so 
			// take it off)
			for(i=0;i<iSubInc;i++)
			{
				Von_mises_PlasticFlow(mStressYield, out mdFdSig);  
				mdGdSig = Gsi.Matrix.Transpose(mdFdSig);

				fDLambda = Gsi.Matrix.Dot(mdFdSig, mExcessStress); 
				
				mTempVec = mCee * mdGdSig ;

				fTemp = Gsi.Matrix.Dot(mdFdSig,mTempVec);  

				fDLambda = fDLambda / fTemp;

				mPlasticStrain = mdFdSig * fDLambda; 

				mPlasticStrainT = Gsi.Matrix.Transpose(mPlasticStrain); 
				
				mPlasticStress = mCee * mPlasticStrainT; 

				mPlasticStressT = Gsi.Matrix.Transpose(mPlasticStress); 

				// reduce actual incremental stresses by plastic amount
				mStressYield += (mExcessStress-mPlasticStressT);

			}
			mUpdatedStress = mStressYield + PWP;
		}
		else
		{
			mUpdatedStress = mStressYield + PWP;
			fLastYield = 0.0;
		}

		//  These are the five values that are returned by this UpdateStresses method.
		//  The solver is expecting valid values for all of these.  These will be used 
		//  with the previous iteration stresses to determine the unbalanced load for
		//  the next iteration.

		Stresses.x = mUpdatedStress[0,0];
		Stresses.y = mUpdatedStress[0,1];
		Stresses.z = mUpdatedStress[0,2];
		Stresses.xyShear = mUpdatedStress[0,3];
		Stresses.voidRatio = 0.0;   //   this is optional to return from this method.
	}
}
