// Sample Add-In a Von Mises yield criteria elastic plastic model. 
//
// This is a Von Mises yield criteria elastic plastic model. It has the CalculateMatrix and
// UpdateStresses methods which MUST be present in an Add-In constitutive model.
//
// It also has an SLOPE/W strength function that returns the undrained strength 
// at the base of a slice based on temperature.
//
// Copyright (C) 2012 GEO-SLOPE International, Ltd.  All rights reserved.
using System;

public class Von_Mises_Thawing : Gsi.Function
{
	//Declare global constants, these will appear in the Key In Materials Dialog
	//for data entry
	public double Cu_frozen;
	public double Cu_thawed;
	public double E;
	public double Poisson;
	public String Freezing_On_Off;
	public String Linked_Analysis_Name;

	// These are used locally in this code and dont require input by the user
	double PWP;
	double Temperature;
	double Cu;

	bool IsTensile;
	int iYieldState; // 0 = elastic, 1=E/P transition, 2 = P on first iteration
	int nIfElastic; 

	static Gsi.DataTable tbls; 
	static System.Object lockThis = new System.Object();
	static int iLastStepLoaded = 0;
 
	System.String sAnalysis_Name;

	public Von_Mises_Thawing()
	{
		iYieldState = 0;
		IsTensile = false;
	}

	// There are four local methods called inside this Von Mises model. They are used
	// as part of the main two methods that required in the Add-In model. Their usage
	// in the code below is highlighted in blue.
	// The first returns a 4x4 matrix for a gauss point with elastic properties only.
	// It takes a modulus and poisson's ratio as input.
	public void CalcElasticCee(double youngsModulus, double Poisson,
		Gsi.Matrix mCee ) // returns mCee matrix
	{
		double fCOM = youngsModulus / ((1.0 + Poisson)*(1.0 - 2.0*Poisson));
		double fCOM1 = 1.0 - Poisson;
		mCee[0,0] = fCOM*fCOM1;
		mCee[0,1] = fCOM*Poisson;
		mCee[0,2] = fCOM*Poisson;
		mCee[0,3] = 0.0;
		mCee[1,0] = fCOM*Poisson;
		mCee[1,1] = fCOM*fCOM1;
		mCee[1,2] = fCOM*Poisson;
		mCee[1,3] = 0.0;
		mCee[2,0] = fCOM*Poisson;
		mCee[2,1] = fCOM*Poisson;
		mCee[2,2] = fCOM*fCOM1;
		mCee[2,3] = 0.0;
		mCee[3,0] = 0.0;
		mCee[3,1] = 0.0;
		mCee[3,2] = 0.0;
		mCee[3,3] = 0.5*youngsModulus/(1.0+Poisson);
	}
	// This returns the stress invariants J2, J3 and Lode Angle and deviatoric stresses. It
	// takes as input x,y,z stress and shear stress..
	public void CalcStressInvariants(Gsi.Matrix mStresses,
		out Gsi.Matrix mDevStress, out double fJ2, out double fJ3,
		out double fTheta, out double fSigMean )
	{
		double fRootJ2;
		double dSin3T;
		double fSigX = mStresses[0,0];
		double fSigY = mStresses[0,1];
		double fSigZ = mStresses[0,2];
		double fTauXY = mStresses[0,3];
		fSigMean = (fSigX+fSigY+fSigZ)/3.0;

		// deviatoric stresses
		mDevStress = mStresses - fSigMean;
		mDevStress[0,3] = mStresses[0,3]; // reset shear stress

		fJ2 = mDevStress[0,3]*mDevStress[0,3] + 0.5*(mDevStress[0,0]*mDevStress[0,0] + mDevStress[0,1]*mDevStress[0,1] + mDevStress[0,2]*mDevStress[0,2]);
		fJ3 = mDevStress[0,2] * (mDevStress[0,2]*mDevStress[0,2] - fJ2);

		fRootJ2 = Math.Pow(fJ2,0.5);

		if(fRootJ2 == 0.0)
			dSin3T = 0.0;
		else
			dSin3T = -3.0*Math.Pow(3.0,0.5)*fJ3/(2.0*fJ2*fRootJ2);

		if(dSin3T > 1.0)
			dSin3T = 1.0;
		if(dSin3T < -1.0)
			dSin3T = -1.0;
		fTheta = Math.Asin(dSin3T) / 3.0;
	}
	// This returns the yield amount stress based on Von Mises. It uses the
	// stress invariant method coded above
	public void Von_mises_Yield(Gsi.Matrix mStresses, out double fYield)
	{
		Gsi.Matrix mDevStress = Gsi.Matrix.Zero(1,4); // Temp Vector
		double fJ2, fJ3, fTheta, fSigMean;
		int nRows = mDevStress.Rows;

		Cu = Cu_thawed;

		if(((Freezing_On_Off == "On") ||(Freezing_On_Off == "on")) && Temperature < 0)  // use frozen Cu for frozen ground
		{
			Cu = Cu_frozen;
		}

		CalcStressInvariants(mStresses, out mDevStress, out fJ2, out fJ3, out fTheta,out fSigMean);
		double fq = Math.Pow(3*fJ2,0.5);
		// Check yield in the post iteration check stress algorithm
		fYield = fq-Math.Pow(3,0.5)*Cu;
	}
	// This returns the Von_Mises plastic flow vector. This also uses the stress
	// invariant method code above.
	public void Von_mises_PlasticFlow( Gsi.Matrix mStresses, out Gsi.Matrix mPlasticFlow)
	{
		double fJ2, fJ3, fTheta, fSigMean;
		Double fTemp;
		Gsi.Matrix mDevStress = Gsi.Matrix.Zero(1,4); // Temp Vector
		CalcStressInvariants(mStresses,out mDevStress,out fJ2,out fJ3, out fTheta, out fSigMean);
		fTemp = 0.5 / Math.Pow(fJ2,0.5);
		mDevStress *= fTemp;
		mDevStress[0,3] *= 2.0;
		fTemp = Math.Pow(3.0,0.5);
		mPlasticFlow = mDevStress*fTemp;
	}

	public void PullBackToYieldSurface(Gsi.Matrix mStressInput, double fYieldStart, out double fYieldNew, out Gsi.Matrix YieldStresses)
	{
		Gsi.Matrix fdFdSij = Gsi.Matrix.Zero(1,4);             //flow vector  = gradient of yield function
		Gsi.Matrix fUnitNormal= Gsi.Matrix.Zero(1,4);
     
		double fAdjustment;          //adjustment factor
		double fMagnitude;           //magnitude of stress vector
		double fFInput;              //input yield function
		double fFNew;
		double fFOld;
		double fTemp;
    
		int nNumIter;
		int i;
		bool nIfDivergent = false;
   
		YieldStresses = Gsi.Matrix.Zero(1,4);

		fFNew = fYieldStart;
		fFInput = fFNew;
		fFOld = fFNew;
		fYieldNew = fYieldStart;

		//Get unit normal to yield surface
		Von_mises_PlasticFlow(mStressInput, out fdFdSij);
         
		UnitStressVector(fdFdSij,out fUnitNormal,out fMagnitude);

		if(fMagnitude != 0)
		{
			//Adjust stress state s.t. it ends up just outside the yield surface
			nNumIter=0;
			fTemp = (double)(-fFInput/100);
			fAdjustment = fTemp;
    
			for(nNumIter=0;nNumIter<100;nNumIter++)
			{
				//store yield function
				fFOld = fFNew;

				//Adjust stresses
				for (i=0;i<4 ;i++ )     
					YieldStresses[0,i] =  mStressInput[0,i]+(fAdjustment*fUnitNormal[0,i]);

				//Check yield function
				Von_mises_Yield(YieldStresses, out fFNew);

				//Check for divergence
				if(Math.Abs(fFNew)>Math.Abs(fFOld))
				{
					nIfDivergent=true;
					break;
				}

				//Check for sign change
				if(fFNew*fFOld < 0.0)
					break;

				fAdjustment += fTemp;
				nNumIter++;      
			}
     
			//if it is divergent, restore starting value
			if(nIfDivergent)
			{
				mStressInput = YieldStresses;
				fFNew = fFInput;
			}
			else
			{
				//Convergence not achieved, restore the initial value if it is closer to the 
				//yield surface
				if(nNumIter > 100)
				{
					if(Math.Abs(fFInput)<Math.Abs(fFNew))
					{
						mStressInput = YieldStresses;
						fFNew = fFInput;
					}     
				}
			}
    
			fYieldNew = fFNew;
		}
	         
	}

	
	public void UnitStressVector(Gsi.Matrix fdFdSij, out Gsi.Matrix fUnitNormal, out double fMagnitude)
	{
		double dSum,dTemp;
		double fTemp;
		int i;

		fUnitNormal = Gsi.Matrix.Zero(1,4);
    
		//Calculate inner product of the input vector
		dSum = 0.0;
		for (i=0;i<4;i++)
		{
			dTemp = fdFdSij[0,i];
        
			if(i>2)
				dTemp = 2.0*(dTemp*dTemp);  //shear components
			else
				dTemp = (dTemp*dTemp);      //normal components
            
			dSum += dTemp;
		}
		dSum = Math.Sqrt(dSum);
		fMagnitude = dSum;
    
		//Calculate unit vector
		fTemp = fMagnitude;
		if(fTemp == 0.0F)
			fTemp = 1.0F;
    
		for (i=0;i<4;i++)
			fUnitNormal[0,i] = fdFdSij[0,i]/fTemp;
	}
	


	// These are the two REQUIRED methods that MUST exist in a user constitutive model.
	// The first returns the [C] matrix that will get put into the EPM at each iteration prior to
	// SOLVING the equations
	// The second computes the new stress state after the incremental displacements and
	// strains are computed in SOLVE at each iteration.
	public void CalculateMatrix( Gsi.Matrix mCee )
	{
		CalcElasticCee(E, Poisson, mCee); // compute elastic [C]

		sAnalysis_Name = Linked_Analysis_Name ; // "3 Transient Thermal";   // for now, must hard wire the name
		
		int nStep;
		int ItNum = (int)GetParam(Gsi.DataParamType.eStepIterationCount);

		if(((Freezing_On_Off == "On") ||(Freezing_On_Off == "on"))) // load table on first slice only
		{
			try
			{
				nStep = (int)GetParam(Gsi.DataParamType.eStepNum);
			}
			catch
			{
				nStep = 1;
			}

			try
			{
				lock(lockThis)
				{
					if ( iLastStepLoaded != nStep )
					{
						tbls = Gsi.DataTable.Load(sAnalysis_Name, Gsi.TableType.eGaussTable,  nStep );
						iLastStepLoaded = nStep;
					}
				}
		
				Temperature = tbls.Get( ID, Gsi.DataParamType.eTemperature) ;
			}
			catch
			{
				Temperature = 0.0; // if no thermal coupling, set this to zero
			}
		}
		
		PWP = GetParam(Gsi.DataParamType.ePWP);
		double Sx = GetParam(Gsi.DataParamType.eXTotalStress) - PWP;
		double Sy = GetParam(Gsi.DataParamType.eYTotalStress) - PWP;
		double Sz = GetParam(Gsi.DataParamType.eZTotalStress) - PWP;
		double Sxy = GetParam(Gsi.DataParamType.eXYShearStress);

		if(ItNum > 1 && iYieldState >= 3)
		{
			Gsi.Matrix mCurrentStress = Gsi.Matrix.Zero(1,4);

			// current solution effective stress state prior to adding the current
			// incremental stress
			mCurrentStress[0,0] = Sx;
			mCurrentStress[0,1] = Sy;
			mCurrentStress[0,2] = Sz;
			mCurrentStress[0,3] = Sxy;

			Gsi.Matrix mdFdSig = Gsi.Matrix.Zero(1,4);
			Gsi.Matrix mdGdSig = Gsi.Matrix.Zero(4,1);
			Gsi.Matrix mTempVRow = Gsi.Matrix.Zero(1,4);
			Gsi.Matrix mTempVCol = Gsi.Matrix.Zero(4,1);
			Gsi.Matrix mCep = Gsi.Matrix.Zero(4,4);

			Von_mises_PlasticFlow(mCurrentStress, out mdFdSig);

			mdGdSig = Gsi.Matrix.Transpose(mdFdSig);
			mTempVCol = mCee * mdGdSig ;
			mTempVRow = Gsi.Matrix.Transpose(mTempVCol);
			double fTemp = Gsi.Matrix.Dot(mdFdSig,mTempVCol);

			if(Math.Abs(fTemp) < 1e-10) 
				fTemp = 1.0e-10;

			fTemp = 1.0/fTemp;

			mCep = Gsi.Matrix.TensorCross(mTempVCol, mTempVRow);

			mCep = mCep * fTemp;

			//Form plastic constitutive matrix.
			mCee = mCee - mCep; // [Cee-Cep]
		}
	}
	public void UpdateStresses( Gsi.Matrix mIncStrain, ref Gsi.Stresses Stresses)
	{
		Gsi.Matrix mIncStress = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mTempStress = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mCee = Gsi.Matrix.Zero(4,4);
		Gsi.Matrix mStressBegin = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mStressNew = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mStressTemp = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mExcessStress = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mPlasticStress = Gsi.Matrix.Zero(4,1);
		Gsi.Matrix mPlasticStressT = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mPlasticStrain = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mPlasticStrainT = Gsi.Matrix.Zero(4,1);
		Gsi.Matrix mdFdSig = Gsi.Matrix.Zero(1,4);
		Gsi.Matrix mdGdSig = Gsi.Matrix.Zero(4,1);
		Gsi.Matrix mTempVec = Gsi.Matrix.Zero(1,4);
		Double fYield = 0.0;
		Double fYieldStart = 0.0;
		Double fReductionFactor = 1.0;
		Double fTemp = 0.0;
		Double fDLambda;
		int i, iSubInc;

		int ItNum = (int)GetParam(Gsi.DataParamType.eStepIterationCount);
		int nYieldFlag_Begin = iYieldState;
		nIfElastic = 0;

		// form the elastic Cee matrix as it is needed for stress calculations
		CalcElasticCee(E, Poisson, mCee);
		
		// compute elastic Ce used to know the stress increment
		// {IncStress} = [Cee]x{IncStrain} This is the full incremental stress
		// based on the elastic load
		
		mIncStress = mIncStrain * mCee;

		mStressBegin[0,0] = GetParam(Gsi.DataParamType.eXTotalStress) - PWP;
		mStressBegin[0,1] = GetParam(Gsi.DataParamType.eYTotalStress) - PWP;
		mStressBegin[0,2] = GetParam(Gsi.DataParamType.eZTotalStress) - PWP;
		mStressBegin[0,3] = GetParam(Gsi.DataParamType.eXYShearStress) ;

		mStressNew = mStressBegin;

		// compute updated stresses with full elastic increment applied.
		// This will be reduced later if there has been yield
		mStressTemp = mStressBegin + mIncStress;
		
		// check yield with starting stresses and then with updated stresses
		Von_mises_Yield(mStressBegin,out fYieldStart);
		Von_mises_Yield(mStressTemp,out fYield);
		
		if(nYieldFlag_Begin >= 3)
			if(fYieldStart < 0)
				fYieldStart = 0;
		
		// find reduction factor
		//---------
		//  Determine the load path of the new stress state for the first iteration:
		//  set nIfElastic flag, YieldFlagNew and calc. reduction factor (fReducFac).
		//  For subsequent iterations, the soil will remain on the same load path as
		//  during the first iteration.
		//---------
		if(ItNum == 1) 
		{            
			//soil has yielded at beginning of iteration
			if(fYieldStart >= 0.0 || nYieldFlag_Begin >=3)        
			{
				//stress state close to or outside yield surface - continue yielding
				if(fYield >= 0.0 || IsTensile )       
				{
					fReductionFactor = 1.0;  
					nIfElastic = 0;     

					if(IsTensile)
						iYieldState = 6;
					else
						iYieldState = 3;

				}
					//stress state is inside yield surface; therefore, elastic
				else                    
				{
					fReductionFactor = 0.0;
					iYieldState = 0;
					nIfElastic = 1;     

				}
			}
				//soil was elastic at beginning of iteration
			else
			{
				if (fYield > 0.0 || IsTensile)   //soil is yielding at the current iteration
				{
					nIfElastic = 0;     

					//Calculate a new reduction factor
					if(fYield*fYieldStart > 0)
						if(fYield > 0)
							fReductionFactor = 1.0;
						else
							fReductionFactor = 0.0;
					else
						fReductionFactor = fYield/(fYield-fYieldStart);  

					if(fReductionFactor > 1)
						fReductionFactor = 1;
					if(fReductionFactor < 0)
						fReductionFactor = 0;
         
					if(IsTensile)
						iYieldState = 6;
					else
						iYieldState = 3;

				}
				else    //soil is elastic
				{
					fReductionFactor = 0.0;
					iYieldState = 0;
					nIfElastic = 1   ;  

				}
			}
		}
		else    //subsequent iterations, soil cannot unload but can become plastic
		{
			if(nYieldFlag_Begin < 3)            //was elastic check for possible yield
			{
				//------
				//  Set yield flag using the new stress state
				//-----
				// soil remains elastic; 
				if( fYield < 0.0 )   //New stresses is within the yield surface  
				{
					fReductionFactor = 0.0;
					iYieldState = 0;
					nIfElastic = 1;     

				}
					// soil has yielded 
				else
				{
					fReductionFactor = 1.0;
					nIfElastic = 0;     


					if(IsTensile)
						iYieldState = 6;
					else
						iYieldState = 3;

					// Yield function at the starting stress state is < 0 indicating that
					// the material has only begun to yield, iterate for Reduction factor.            
					if (fYieldStart < 0.0)
					{
						if(fYield*fYieldStart > 0)
							if(fYield > 0)
								fReductionFactor = 1.0;
							else
								fReductionFactor = 0.0;
						else
							fReductionFactor = fYield/(fYield-fYieldStart);  

						if(fReductionFactor > 1)
							fReductionFactor = 1;
						if(fReductionFactor < 0)
							fReductionFactor = 0;

					}
				}
			}
			else    //continue yielding - YieldFlag is unchanged
			{
				fReductionFactor = 1.0;
				nIfElastic = 0;     


				//29Jun94 Ensure that the yield function values reflect the yielded state
				if(fYieldStart < 0.0)
					fYieldStart = 0.0;
				if(fYield < 0.001)
					fYield = 0.001;

				if(IsTensile)
					iYieldState = 6;
				else
					iYieldState = 3;
			}               
		}



		if(1.0 - fReductionFactor > 0.0)
		{
			mStressNew = mStressBegin + mIncStress*(1-fReductionFactor);
		}

		iSubInc = 0;


		if(nIfElastic == 0 || fYield > 0.1)
		{
			// find plastic strain for the given total strain increment, compute the
			// elastic stress that would have caused this strain,
			// then reduce the initially computed stress increment by this plastic
			//stress amount (there can be no change in stress for plastic strain so
			// take it off)

			iSubInc = (int)(Math.Abs(fYield));
			iSubInc += 1;

			if(iSubInc > 32)
				iSubInc = 32; // max same as internal code.

			mExcessStress = mIncStress*fReductionFactor/iSubInc;

			for(i=0;i<iSubInc;i++)
			{
				Von_mises_PlasticFlow(mStressNew, out mdFdSig);
				mdGdSig = Gsi.Matrix.Transpose(mdFdSig);
				fDLambda = Gsi.Matrix.Dot(mdFdSig, mExcessStress);
				mTempVec = mCee * mdGdSig;
				fTemp = Gsi.Matrix.Dot(mdFdSig,mTempVec);
				fDLambda = fDLambda / fTemp;
				mPlasticStrain = mdFdSig * fDLambda;
				mPlasticStrainT = Gsi.Matrix.Transpose(mPlasticStrain);
				mPlasticStress = mCee * mPlasticStrainT;
				mPlasticStressT = Gsi.Matrix.Transpose(mPlasticStress);
				// reduce actual incremental stresses by plastic amount
				mStressNew += (mExcessStress-mPlasticStressT);
			}
			//Check yield at the new stress state
			double fTestYield;
			Von_mises_Yield(mStressNew,out fTestYield);

			//Pull-back function here if fFNew > gfFuzz_Yield
			if(Math.Abs(fTestYield)>0.1)   
			{
				double fYieldUpdate;
				Gsi.Matrix mYieldSurfaceStresses = Gsi.Matrix.Zero(1,4);
				PullBackToYieldSurface(mStressNew, fTestYield, out fYieldUpdate, out mYieldSurfaceStresses);
				mStressNew = mYieldSurfaceStresses;
				//	fYield = fYieldUpdate;
			} 
		}

		// These are the five values that are returned by this UpdateStresses method.
		// The solver is expecting valid values for all of these. These will be used
		// with the previous iteration stresses to determine the unbalanced load for
		// the next iteration.
		Stresses.x = mStressNew[0,0];
		Stresses.y = mStressNew[0,1];
		Stresses.z = mStressNew[0,2];
		Stresses.xyShear = mStressNew[0,3];
		Stresses.voidRatio = Cu; // Temperature; // this is optional to return from this method.

		SetParam(Gsi.DataParamType.eCustomParam1, Cu);

		if((Freezing_On_Off == "On") ||(Freezing_On_Off == "on"))
			SetParam(Gsi.DataParamType.eCustomParam2, Temperature);
	}
}



public class Cohesion_Freeze_Thaw : Gsi.Function
{
	public double Cu_frozen;
	public double Cu_thawed;
	public String Freezing_On_Off;
	public String Linked_Analysis_Name;
	public double Test_temperture;

	double Temperature;

	static Gsi.DataTable tbls; // = new Gsi.DataTable();  
	static System.Object lockThis = new System.Object();
	static int iLastStepLoaded = 0;
 
	System.String sAnalysis_Name;

	public double Calculate( double Passed_in_X ) 
	{
		// read in Y-coordinate at slice base center 
		double BaseCenterY;
		double BaseCenterX;
		int nStep;

		try
		{
			BaseCenterY = GetParam(Gsi.DataParamType.eYCoord);  // this will get live data from the solver.
			BaseCenterX = GetParam(Gsi.DataParamType.eXCoord);  // this will get live data from the solver.
		}
		catch
		{
			BaseCenterY = 0;
			BaseCenterX = Passed_in_X;
		}

		sAnalysis_Name = Linked_Analysis_Name ; // "3 Transient Thermal";   // for now, must hard wire the name

		if(((Freezing_On_Off == "On") ||(Freezing_On_Off == "on"))) // load table on first slice only
		{
			try
			{
				nStep = (int)GetParam(Gsi.DataParamType.eStepNum);
			}
			catch
			{
				nStep = iLastStepLoaded;
			}

			try
			{
				lock(lockThis)
				{
					if ( iLastStepLoaded != nStep )
					{
						tbls = Gsi.DataTable.Load(sAnalysis_Name, Gsi.TableType.eNodeTable,  nStep );
						iLastStepLoaded = nStep;
					}
				}
			}
			catch
			{
			}

			if(nStep != 0) // so solver is calling this
				Temperature = Gsi.Mesh.ComputeNodalValueAtXY( tbls, BaseCenterX, BaseCenterY , Gsi.DataParamType.eTemperature );
			else
				Temperature = Test_temperture; // define is calling it for graphing purposes
		}

		if(((Freezing_On_Off == "On") ||(Freezing_On_Off == "on")) && Temperature < 0)  // use frozen Cu for frozen ground
		{
			return (Cu_frozen);
		}
		else
            return(Cu_thawed);
	}
}