///////////////////////////////////////////////////////////////////////////////
// Slip Surface Strength.cs
// Copyright (C) 2007 GEO-SLOPE International, Ltd.  All rights reserved.
//
// Notes about coding Strength Function AddIn for SLOPE/W:
//
//	You can change the names of functions and variables EXCEPT the function 
//  called "Calculate(double x)" which MUST exist inside the Class.
// 
//	Variables you want to see in the KeyIn Function dialog should be "public" 
//  in the Class
//
//  An instance of Strength Function class exists for every slip surface slice 
//  in the solver. The Calculate routine receives the effective base normal 
//  stress at the base of the current slice as its argument.
//
//	Here are the "live" data names you can ask for in a SLOPE/W 
//  GetParam(Gsi.DataParamType.eXCoord) statement.  The names below would appear 
//  in place of the ".eXCoord" words in the above line. An example of this 
//  statement is used in the sample function below.
//
//		eXCoord     	- the x coordinate of the base of the slice
//		eYCoord	    	- the y coordinate of the base of the slice
//		ePWP	    	- the pore-water pressure at the base of the slice. 
//		eBaseAngle  	- the base slice angle in degrees (note the -ve angle and 
//                        +ve angle convention)
// 		eBaseNormalStress - the total normal stress at the base of a slice
///////////////////////////////////////////////////////////////////////////////
using System;

///////////////////////////////////////////////////////////////////////////////
// Strength_Based_On_Position
//
// This function allows users to specify c and phi on either side of an arbitrary 
// input x coordinate 
// 
// Inputs:
// 		X_Critical 
// 		c_Left
// 		phi_Left
// 		c_Right
// 		phi_Right
//      Assume_X_Coordinate_for_viewing
// 
// Output:
//      Shear strength at base center of a slice
//
///////////////////////////////////////////////////////////////////////////////

// This is the first Class in this file.  You can have more than one.  The name 
// is "Strength_Based_On_Position" which you can change. This is the same name 
// that will be given to the AddIn *.dll file you will compile and save in the 
// GeoStudio AddIns folder.

// The : Gsi.Function text after the name is required if you want to ask for 
// "live" data from the solver via the GetParam routine.  The allowable "live"  
// data variables are given above.

public class Strength_Based_On_Position : Gsi.Function
{
    // Here is the list of variables you want to enter using the KeyIn Function 
    // dialog.
    public double X_Critical;
    public double c_Left;
    public double phi_Left;
    public double c_Right;
    public double phi_Right;

    // Normally this function reads the X-coordinate of the current slice in the 
    // solver, but if you want to view the graph in the KeyIn Function dialog, 
    // you can enter an arbitrary X coordinate of a slice used for graphing.
    public double Assume_X_Coordinate_for_viewing;

    // This is the main Function calculator.  You CANNOT change the name. You can 
    // change the variable name(s). EffectiveBaseNormalStress for the Functions
    // slice is passed in by the solver.  You can use it and / or get other "live" 
    // data values associated with the Functions slice from the solver while its 
    // solving.
    public double Calculate(double EffectiveBaseNormalStress)
    {
        // These are local variables used in the calcuation of the value this 
        // function will return.		
        double BaseCenterX;
        double Degree_to_Radian = 3.141592654 / 180.0;
        double Strength;

        // This "try" and "catch" phrase is used to specify what to do if 
        // a program statement fails. GetParam fails when run in the KeyIn Functions 
        // dialog for graphing. Using try/catch, a work around value can be specified
        // for BaseCenterX in KeyIn Functions dialog so the function can be graphed. 
        try
        {
            // First try to get live data from the solver, in this case, the x 
            // coordinate of the slice.
            BaseCenterX = GetParam(Gsi.DataParamType.eXCoord);  
        }
        catch
        {
            // If GetParam fails to find the x-coordinate of the slice, assume
            // this Function is being called from the KeyIn Functions dialog
            // for graphing and use the Assume_X_Coordinate_for_viewing.
            BaseCenterX = Assume_X_Coordinate_for_viewing;	
        }

        // Check if slice x is to the left or to the right of the critical X
        if (BaseCenterX < X_Critical)
        {
            // Mohr Coulomb strength at slice base 
            Strength = c_Left + EffectiveBaseNormalStress * Math.Tan(phi_Left * Degree_to_Radian);
        }
        else
        {
            // Mohr Coulomb strength at slice base 
            Strength = c_Right + EffectiveBaseNormalStress * Math.Tan(phi_Right * Degree_to_Radian);
        }

        // return the computed Strength value

        return (Strength);

    }  // this is the end of the calculate function

} // this is the end of the first Class in this file.


///////////////////////////////////////////////////////////////////////////////
// Strength_Mohr_Columb
//
// This function allows users to specify c and phi and compute strength at the 
// slice base using a Mohr Columb soil model 
//  
// Inputs:
// 		c
// 		Phi
// 
// Output:
//      Shear strength at base center of a slice
//
// Developed by:  Leonard Lam May 23, 2007
///////////////////////////////////////////////////////////////////////////////
public class Strength_Mohr_Columb 
{
	public double c;
	public double phi;

	// this is the main function calculator.  You CANNOT change the name. You 
    // can change the variable name.
	public double Calculate( double EffectiveBaseNormalStress )   
	{
		double Degree_to_Radian =  3.141592654 / 180.0;

		// Mohr Coulomb strength at slice base 
		double Strength =  c + EffectiveBaseNormalStress * Math.Tan( phi * Degree_to_Radian );
      
		// return the function Strength value
		return(Strength);
	}
}


///////////////////////////////////////////////////////////////////////////////
// Strength_Between_X_Coordinates
//
// This function allows users to specify c and phi on the left and right 
// locations of a slope, c and phi are assumed to vary linearly along the left 
// and right locations
// 
// Inputs:
// 		X-Coordinate Left 
// 		C Left
// 		Phi Left
// 		X-Coordinate Right
// 		C Right
// 		Phi Right
// 
// Output:
//      Shear strength at base center of a slice
//
// Developed by:  Leonard Lam May 23, 2007
///////////////////////////////////////////////////////////////////////////////
public class Strength_Between_X_Coordinates : Gsi.Function
{
	public double X_Coordinate_Left;
	public double c_Left;
	public double phi_Left;

	public double X_Coordinate_Right;
	public double c_Right;
	public double phi_Right;

	// Normally the solver will pass the slice X-coordinate to the function, but 
	// if you want to view the graph in DEFINE, you can enter an arbitrary X 
    // coordinate of a slice
	public double Assume_X_Coordinate_for_viewing;

	double fCInt,fPhiInt,fCSlope,fPhiSlope;

	// this is the main function calculator.  You CANNOT change the name. You can change the variable name.
	public double Calculate( double EffectiveBaseNormalStress )   
	{
		double BaseCenterX;
		double c_New;
		double phi_New;
		double Degree_to_Radian =  3.141592654 / 180.0;

		// read in X-coordiate at slice base center 
		try
		{
			BaseCenterX = GetParam(Gsi.DataParamType.eXCoord);  // this will get live data from the solver.
		}
		catch
		{
			BaseCenterX = Assume_X_Coordinate_for_viewing;
		}

		// compute the linear c equation
		fCSlope = (c_Right - c_Left) / (X_Coordinate_Right - X_Coordinate_Left);
		fPhiSlope = (phi_Right - phi_Left) / (X_Coordinate_Right - X_Coordinate_Left);
		
		fCInt = c_Left - fCSlope * X_Coordinate_Left;
		fPhiInt = phi_Left - fPhiSlope * X_Coordinate_Left;
		
		c_New = fCInt + fCSlope * BaseCenterX;
		phi_New = fPhiInt + fPhiSlope * BaseCenterX;

		// Mohr Coulomb strength at slice base 
		double Strength =  c_New + EffectiveBaseNormalStress * Math.Tan( phi_New * Degree_to_Radian );
			
		// return the function Strength value
		return(Strength);
	}
}

///////////////////////////////////////////////////////////////////////////////
//Strength_Between_Y_Coordinates
//
// This function allows users to specify c and phi on the top and bottom 
// locations of a slope, c and phi are assumed to vary linearly along the top 
// and bottom locations
// 
// Inputs:
// 		X-Coordinate Top 
// 		C Top
// 		Phi Top
// 		X-Coordinate Bottom
// 		C Bottom
// 		Phi Bottom
// 
// Output:
//      Shear strength at base center of a slice
//
// Developed by:  Leonard Lam May 23, 2007
///////////////////////////////////////////////////////////////////////////////

public class Strength_Between_Y_Coordinates : Gsi.Function
{
	public double Y_Coordinate_Top;
	public double c_Top;
	public double phi_Top;

	public double Y_Coordinate_Bottom;
	public double c_Bottom;
	public double phi_Bottom;

	// Normally the solver will pass the slice Y-coordinate to the function, but 
	// if you want to view the graph in DEFINE, you can enter an arbitrary Y 
    // coordinate of a slice
	public double Assume_Y_Coordinate_for_viewing;

	// this is the main function calculator.  You CANNOT change the name. You can 
    // change the variable name.
	public double Calculate( double EffectiveBaseNormalStress ) 
	{
		double Degree_to_Radian =  3.141592654 / 180.0;

		// read in Y-coordiate at slice base center 
        double BaseCenterY;
        try
		{
			BaseCenterY = GetParam(Gsi.DataParamType.eYCoord);  // this will get live data from the solver.
		}
		catch
		{
			BaseCenterY = Assume_Y_Coordinate_for_viewing;
		}

		// compute the linear c equation
		double fCSlope = (c_Top - c_Bottom) / (Y_Coordinate_Top - Y_Coordinate_Bottom);
		double fPhiSlope = (phi_Top - phi_Bottom) / (Y_Coordinate_Top - Y_Coordinate_Bottom);
		
		double fCInt = c_Bottom - fCSlope * Y_Coordinate_Bottom;
		double fPhiInt = phi_Bottom - fPhiSlope * Y_Coordinate_Bottom;
		
		double c_New = fCInt + fCSlope * BaseCenterY;
		double phi_New = fPhiInt + fPhiSlope * BaseCenterY;
        SetParam(Gsi.DataParamType.eCustomParam1, c_New);   //Write computed C to file for graphing
        SetParam(Gsi.DataParamType.eCustomParam2, phi_New); //Write computed phi to file for graphing

		// Mohr Coulomb strength at slice base 
		double Strength =  c_New + EffectiveBaseNormalStress * Math.Tan( phi_New * Degree_to_Radian );
		
		// return the function Strength value
		return(Strength);
	}
}

